import json
import logging
import re

from dagster.core.definitions import SolidHandle
from dagster.core.events import DagsterEvent
from dagster.core.execution_plan.objects import StepFailureData
from dagster.core.log import DAGSTER_DEFAULT_LOGGER, DagsterLog
from dagster.utils.error import SerializableErrorInfo

REGEX_UUID = r'[a-z-0-9]{8}\-[a-z-0-9]{4}\-[a-z-0-9]{4}\-[a-z-0-9]{4}\-[a-z-0-9]{12}'
REGEX_TS = r'\d{4}\-\d{2}\-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}'


def _setup_logger(name):
    class TestLogger(logging.Logger):
        def __init__(self, name):
            super(TestLogger, self).__init__(name)

    logger = TestLogger(name)

    captured_results = []

    def log_fn(msg, *args, **kwargs):  # pylint:disable=unused-argument
        captured_results.append(msg)

    for level in ['debug', 'info', 'warning', 'error', 'critical']:
        setattr(logger, level, log_fn)

    return captured_results, logger


def _regex_match_kv_pair(regex, kv_pairs):
    return any([re.match(regex, kv_pair) for kv_pair in kv_pairs])


def _validate_basic(kv_pairs):
    assert 'orig_message="test"' in kv_pairs
    assert _regex_match_kv_pair(r'log_message_id="{0}"'.format(REGEX_UUID), kv_pairs)
    assert _regex_match_kv_pair(r'log_timestamp="{0}"'.format(REGEX_TS), kv_pairs)


def test_logging_basic():
    captured_results, logger = _setup_logger('test')

    dl = DagsterLog('123', {}, [logger])
    dl.info('test')

    kv_pairs = set(captured_results[0].strip().split())
    _validate_basic(kv_pairs)


def test_multiline_logging_basic():
    captured_results, logger = _setup_logger(DAGSTER_DEFAULT_LOGGER)

    dl = DagsterLog('123', {}, [logger])
    dl.info('test')

    kv_pairs = captured_results[0].replace(' ', '').split('\n')[1:]
    _validate_basic(kv_pairs)


def test_multiline_logging_complex():
    msg = 'DagsterEventType.STEP_FAILURE for step start.materialization.output.result.0'
    kwargs = {
        'pipeline': 'example',
        'pipeline_name': 'example',
        'step_key': 'start.materialization.output.result.0',
        'solid': 'start',
        'solid_definition': 'emit_num',
        'dagster_event': DagsterEvent(
            event_type_value='STEP_FAILURE',
            pipeline_name='error_monster',
            step_key='start.materialization.output.result.0',
            solid_handle=SolidHandle('start', 'emit_num'),
            step_kind_value='MATERIALIZATION_THUNK',
            logging_tags={
                'pipeline': 'error_monster',
                'step_key': 'start.materialization.output.result.0',
                'solid': 'start',
                'solid_definition': 'emit_num',
            },
            event_specific_data=StepFailureData(
                error=SerializableErrorInfo(
                    message="FileNotFoundError: [Errno 2] No such file or directory: '/path/to/file'\n",
                    stack=[
                        '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/errors.py", line 186, in user_code_error_boundary\n    yield\n',
                        '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/execution_plan/simple_engine.py", line 365, in _iterate_step_outputs_within_boundary\n    for step_output in gen:\n',
                        '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/execution_plan/materialization_thunk.py", line 28, in _fn\n    runtime_type.output_schema.materialize_runtime_value(config_spec, runtime_value)\n',
                        '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/types/config_schema.py", line 93, in materialize_runtime_value\n    return func(config_value, runtime_value)\n',
                        '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/types/config_schema.py", line 110, in _selector\n    return func(selector_key, selector_value, runtime_value)\n',
                        '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/types/builtin_config_schemas.py", line 59, in _builtin_output_schema\n    with open(json_file_path, \'w\') as ff:\n',
                    ],
                    cls_name='FileNotFoundError',
                )
            ),
        ),
    }

    captured_results, logger = _setup_logger(DAGSTER_DEFAULT_LOGGER)

    dl = DagsterLog('123', {}, [logger])
    dl.info(msg, **kwargs)

    kv_pairs = set(captured_results[0].split('\n')[1:])

    expected_pairs = [
        '        orig_message = "DagsterEventType.STEP_FAILURE for step start.materialization.output.result.0"',
        '              run_id = "123"',
        '            pipeline = "example"',
        '    solid_definition = "emit_num"',
        '       pipeline_name = "example"',
        '               solid = "start"',
        '            step_key = "start.materialization.output.result.0"',
    ]
    for e in expected_pairs:
        assert e in kv_pairs

    assert _regex_match_kv_pair(r'      log_message_id = "{0}"'.format(REGEX_UUID), kv_pairs)
    assert _regex_match_kv_pair(r'       log_timestamp = "{0}"'.format(REGEX_TS), kv_pairs)

    expected_dagster_event = {
        'event_specific_data': [
            [
                "FileNotFoundError: [Errno 2] No such file or directory: '/path/to/file'\n",
                [
                    '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/errors.py", line 186, in user_code_error_boundary\n    yield\n',
                    '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/execution_plan/simple_engine.py", line 365, in _iterate_step_outputs_within_boundary\n    for step_output in gen:\n',
                    '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/execution_plan/materialization_thunk.py", line 28, in _fn\n    runtime_type.output_schema.materialize_runtime_value(config_spec, runtime_value)\n',
                    '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/types/config_schema.py", line 93, in materialize_runtime_value\n    return func(config_value, runtime_value)\n',
                    '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/types/config_schema.py", line 110, in _selector\n    return func(selector_key, selector_value, runtime_value)\n',
                    '  File "/Users/nate/src/dagster/python_modules/dagster/dagster/core/types/builtin_config_schemas.py", line 59, in _builtin_output_schema\n    with open(json_file_path, \'w\') as ff:\n',
                ],
                'FileNotFoundError',
            ]
        ],
        'event_type_value': 'STEP_FAILURE',
        'pipeline_name': 'error_monster',
        'solid_handle': ['start', 'emit_num', None],
        'step_key': 'start.materialization.output.result.0',
        'step_kind_value': 'MATERIALIZATION_THUNK',
        'logging_tags': {
            'pipeline': 'error_monster',
            'solid': 'start',
            'solid_definition': 'emit_num',
            'step_key': 'start.materialization.output.result.0',
        },
    }
    dagster_event = json.loads(
        [pair for pair in kv_pairs if 'dagster_event' in pair][0].strip('       dagster_event = ')
    )
    assert dagster_event == expected_dagster_event



from abc import ABCMeta, abstractmethod
from collections import namedtuple

import six

from dagster import check

from dagster.core.execution.execution_context import SystemPipelineExecutionContext
from .object_store import ObjectStore
from .runs import RunStorageMode
from .types.runtime import RuntimeType


class StepOutputHandle(namedtuple('_StepOutputHandle', 'step_key output_name')):
    @staticmethod
    def from_step(step, output_name='result'):
        from .execution_plan.objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)

        return StepOutputHandle(step.key, output_name)

    def __new__(cls, step_key, output_name='result'):
        return super(StepOutputHandle, cls).__new__(
            cls,
            step_key=check.str_param(step_key, 'step_key'),
            output_name=check.str_param(output_name, 'output_name'),
        )


class IntermediatesManager(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def get_intermediate(self, context, runtime_type, step_output_handle):
        pass

    @abstractmethod
    def set_intermediate(self, context, runtime_type, step_output_handle, value):
        pass

    @abstractmethod
    def has_intermediate(self, context, step_output_handle):
        pass

    @abstractmethod
    def copy_intermediate_from_prev_run(self, context, previous_run_id, step_output_handle):
        pass

    def all_inputs_covered(self, context, step):
        return len(self.uncovered_inputs(context, step)) == 0

    def uncovered_inputs(self, context, step):
        from .execution_plan.objects import ExecutionStep

        check.inst_param(step, 'step', ExecutionStep)
        uncovered_inputs = []
        for step_input in step.step_inputs:
            if not self.has_intermediate(context, step_input.prev_output_handle):
                uncovered_inputs.append(step_input.prev_output_handle)
        return uncovered_inputs


class InMemoryIntermediatesManager(IntermediatesManager):
    def __init__(self):

        self.values = {}
        self.storage_mode = RunStorageMode.IN_MEMORY

    # Note:
    # For the in-memory manager context and runtime are currently optional
    # because they are not strictly required. So we allow one to access
    # these values in places where those are not immediately avaiable
    # but one wants to inspect intermediates. This is useful in test contexts
    # especially

    def get_intermediate(self, context, runtime_type, step_output_handle):
        check.opt_inst_param(context, 'context', SystemPipelineExecutionContext)
        check.opt_inst_param(runtime_type, 'runtime_type', RuntimeType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        return self.values[step_output_handle]

    def set_intermediate(self, context, runtime_type, step_output_handle, value):
        check.opt_inst_param(context, 'context', SystemPipelineExecutionContext)
        check.opt_inst_param(runtime_type, 'runtime_type', RuntimeType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        self.values[step_output_handle] = value

    def has_intermediate(self, context, step_output_handle):
        check.opt_inst_param(context, 'context', SystemPipelineExecutionContext)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        return step_output_handle in self.values

    def copy_intermediate_from_prev_run(self, context, previous_run_id, step_output_handle):
        check.failed('not implemented in in memory')


class ObjectStoreIntermediatesManager(IntermediatesManager):
    def __init__(self, object_store):
        self._object_store = check.inst_param(object_store, 'object_store', ObjectStore)
        self.storage_mode = self._object_store.storage_mode

    def _get_paths(self, step_output_handle):
        return ['intermediates', step_output_handle.step_key, step_output_handle.output_name]

    def get_intermediate(self, context, runtime_type, step_output_handle):
        check.inst_param(context, 'context', SystemPipelineExecutionContext)
        check.inst_param(runtime_type, 'runtime_type', RuntimeType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        check.invariant(self.has_intermediate(context, step_output_handle))

        return self._object_store.get_value(
            context=context, runtime_type=runtime_type, paths=self._get_paths(step_output_handle)
        )

    def set_intermediate(self, context, runtime_type, step_output_handle, value):
        check.inst_param(context, 'context', SystemPipelineExecutionContext)
        check.inst_param(runtime_type, 'runtime_type', RuntimeType)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)
        check.invariant(not self.has_intermediate(context, step_output_handle))

        return self._object_store.set_value(
            obj=value,
            context=context,
            runtime_type=runtime_type,
            paths=self._get_paths(step_output_handle),
        )

    def has_intermediate(self, context, step_output_handle):
        check.inst_param(context, 'context', SystemPipelineExecutionContext)
        check.inst_param(step_output_handle, 'step_output_handle', StepOutputHandle)

        return self._object_store.has_object(context, self._get_paths(step_output_handle))

    def copy_intermediate_from_prev_run(self, context, previous_run_id, step_output_handle):
        return self._object_store.copy_object_from_prev_run(
            context, previous_run_id, self._get_paths(step_output_handle)
        )



import datetime
import itertools
import logging
import uuid

from dagster import check, seven

DAGSTER_META_KEY = 'dagster_meta'
DAGSTER_DEFAULT_LOGGER = 'dagster'


def _dump_value(value):
    # dump namedtuples as objects instead of arrays
    if isinstance(value, tuple) and hasattr(value, '_asdict'):
        return seven.json.dumps(value._asdict())

    return seven.json.dumps(value)


def _kv_message(all_items, multiline=False):
    sep = '\n' if multiline else ' '
    format_str = '{key:>20} = {value}' if multiline else '{key}={value}'
    return sep + sep.join(
        [format_str.format(key=key, value=_dump_value(value)) for key, value in all_items]
    )


class DagsterLog:
    def __init__(self, run_id, logging_tags, loggers):
        self.run_id = check.str_param(run_id, 'run_id')
        self.logging_tags = check.dict_param(logging_tags, 'logging_tags')
        self.loggers = check.list_param(loggers, 'loggers', of_type=logging.Logger)

    def _log(self, method, orig_message, message_props):
        check.str_param(method, 'method')
        check.str_param(orig_message, 'orig_message')
        check.dict_param(message_props, 'message_props')

        check.invariant(
            'extra' not in message_props, 'do not allow until explicit support is handled'
        )
        check.invariant(
            'exc_info' not in message_props, 'do not allow until explicit support is handled'
        )

        check.invariant('orig_message' not in message_props, 'orig_message reserved value')
        check.invariant('message' not in message_props, 'message reserved value')
        check.invariant('log_message_id' not in message_props, 'log_message_id reserved value')
        check.invariant('log_timestamp' not in message_props, 'log_timestamp reserved value')

        log_message_id = str(uuid.uuid4())

        log_timestamp = datetime.datetime.utcnow().isoformat()

        synth_props = {
            'orig_message': orig_message,
            'log_message_id': log_message_id,
            'log_timestamp': log_timestamp,
            'run_id': self.run_id,
        }

        # We first generate all props for the purpose of producing the semi-structured
        # log message via _kv_messsage
        all_props = dict(
            itertools.chain(synth_props.items(), self.logging_tags.items(), message_props.items())
        )

        msg_with_structured_props = _kv_message(all_props.items())
        msg_with_multiline_structured_props = _kv_message(all_props.items(), multiline=True)

        # So here we use the arbitrary key DAGSTER_META_KEY to store a dictionary of
        # all the meta information that dagster injects into log message.
        # The python logging module, in its infinite wisdom, actually takes all the
        # keys in extra and unconditionally smashes them into the internal dictionary
        # of the logging.LogRecord class. We used a reserved key here to avoid naming
        # collisions with internal variables of the LogRecord class.
        # See __init__.py:363 (makeLogRecord) in the python 3.6 logging module source
        # for the gory details.
        # getattr(self.logger, method)(
        #     message_with_structured_props, extra={DAGSTER_META_KEY: all_props}
        # )

        for logger in self.loggers:
            logger_method = check.is_callable(getattr(logger, method))
            if logger.name == DAGSTER_DEFAULT_LOGGER:
                logger_method(
                    msg_with_multiline_structured_props, extra={DAGSTER_META_KEY: all_props}
                )
            else:
                logger_method(msg_with_structured_props, extra={DAGSTER_META_KEY: all_props})

    def debug(self, msg, **kwargs):
        '''
        Debug level logging directive. Ends up invoking loggers with DEBUG error level.

        The message will be automatically adorned with context information about the name
        of the pipeline, the name of the solid, and so forth. The use can also add
        context values during execution using the value() method of ExecutionContext.
        Therefore it is generally unnecessary to include this type of information
        (solid name, pipeline name, etc) in the log message unless it is critical
        for the readability/fluency of the log message text itself.

        You can optionally additional context key-value pairs to an individual log
        message using the keyword args to this message

        Args:
            msg (str): The core string
            **kwargs (Dict[str, Any]): Additional context values for only this log message.
        '''
        return self._log('debug', msg, kwargs)

    def info(self, msg, **kwargs):
        '''Log at INFO level

        See debug()'''
        return self._log('info', msg, kwargs)

    def warning(self, msg, **kwargs):
        '''Log at WARNING level

        See debug()'''
        return self._log('warning', msg, kwargs)

    def error(self, msg, **kwargs):
        '''Log at ERROR level

        See debug()'''
        return self._log('error', msg, kwargs)

    def critical(self, msg, **kwargs):
        '''Log at CRITICAL level

        See debug()'''
        return self._log('critical', msg, kwargs)



from collections import namedtuple

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.types.config import ConfigType, ALL_CONFIG_BUILTINS
from dagster.core.types.iterate_types import iterate_config_types

from .context import PipelineContextDefinition
from .environment_configs import define_environment_cls, EnvironmentClassCreationData
from .pipeline import PipelineDefinition
from .pipeline_creation import construct_runtime_type_dictionary
from .solid import SolidDefinition


def iterate_solid_def_types(solid_def):
    if solid_def.config_field:
        for runtime_type in iterate_config_types(solid_def.config_field.config_type):
            yield runtime_type


def _gather_all_schemas(solid_defs):
    runtime_types = construct_runtime_type_dictionary(solid_defs)
    for rtt in runtime_types.values():
        if rtt.input_schema:
            for ct in iterate_config_types(rtt.input_schema.schema_type):
                yield ct
        if rtt.output_schema:
            for ct in iterate_config_types(rtt.output_schema.schema_type):
                yield ct


def _gather_all_config_types(solid_defs, context_definitions, environment_type):
    check.list_param(solid_defs, 'solid_defs', SolidDefinition)
    check.dict_param(
        context_definitions,
        'context_definitions',
        key_type=str,
        value_type=PipelineContextDefinition,
    )

    check.inst_param(environment_type, 'environment_type', ConfigType)

    for solid_def in solid_defs:
        for runtime_type in iterate_solid_def_types(solid_def):
            yield runtime_type

    for context_definition in context_definitions.values():
        if context_definition.config_field:
            context_config_type = context_definition.config_field.config_type
            for runtime_type in iterate_config_types(context_config_type):
                yield runtime_type

    for runtime_type in iterate_config_types(environment_type):
        yield runtime_type


def construct_config_type_dictionary(solid_defs, context_definitions, environment_type):
    check.list_param(solid_defs, 'solid_defs', SolidDefinition)
    check.dict_param(
        context_definitions,
        'context_definitions',
        key_type=str,
        value_type=PipelineContextDefinition,
    )
    check.inst_param(environment_type, 'environment_type', ConfigType)

    type_dict_by_name = {t.name: t for t in ALL_CONFIG_BUILTINS}
    type_dict_by_key = {t.key: t for t in ALL_CONFIG_BUILTINS}
    all_types = list(
        _gather_all_config_types(solid_defs, context_definitions, environment_type)
    ) + list(_gather_all_schemas(solid_defs))

    for config_type in all_types:
        name = config_type.name
        if name and name in type_dict_by_name:
            if type(config_type) is not type(type_dict_by_name[name]):
                raise DagsterInvalidDefinitionError(
                    (
                        'Type names must be unique. You have constructed two different '
                        'instances of types with the same name "{name}".'
                    ).format(name=name)
                )
        elif name:
            type_dict_by_name[config_type.name] = config_type

        key = config_type.key

        if key in type_dict_by_key:
            if type(config_type) is not type(type_dict_by_key[key]):
                raise DagsterInvalidDefinitionError(
                    (
                        'Type keys must be unique. You have constructed two different '
                        'instances of types with the same key "{key}".'
                    ).format(key=key)
                )

        else:
            type_dict_by_key[config_type.key] = config_type

    return type_dict_by_name, type_dict_by_key


class EnvironmentSchema(
    namedtuple(
        '_EnvironmentSchema', 'environment_type config_type_dict_by_name config_type_dict_by_key'
    )
):
    def __new__(cls, environment_type, config_type_dict_by_name, config_type_dict_by_key):
        return super(EnvironmentSchema, cls).__new__(
            cls,
            environment_type=check.inst_param(environment_type, 'environment_type', ConfigType),
            config_type_dict_by_name=check.dict_param(
                config_type_dict_by_name,
                'config_type_dict_by_name',
                key_type=str,
                value_type=ConfigType,
            ),
            config_type_dict_by_key=check.dict_param(
                config_type_dict_by_key,
                'config_type_dict_by_key',
                key_type=str,
                value_type=ConfigType,
            ),
        )

    def has_config_type(self, name):
        check.str_param(name, 'name')
        return name in self.config_type_dict_by_name

    def config_type_named(self, name):
        check.str_param(name, 'name')
        return self.config_type_dict_by_name[name]

    def config_type_keyed(self, key):
        check.str_param(key, 'key')
        return self.config_type_dict_by_key[key]

    def all_config_types(self):
        return self.config_type_dict_by_key.values()


def create_environment_schema(pipeline_def, mode=None):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.opt_str_param(mode, 'mode')

    mode_definition = pipeline_def.get_mode_definition(mode)

    environment_cls = define_environment_cls(
        EnvironmentClassCreationData(
            pipeline_def.name,
            pipeline_def.solids,
            pipeline_def.context_definitions if mode_definition is None else None,
            pipeline_def.dependency_structure,
            mode_definition,
        )
    )

    environment_type = environment_cls.inst()

    config_type_dict_by_name, config_type_dict_by_key = construct_config_type_dictionary(
        pipeline_def.solid_defs, pipeline_def.context_definitions, environment_type
    )

    return EnvironmentSchema(
        environment_type=environment_type,
        config_type_dict_by_name=config_type_dict_by_name,
        config_type_dict_by_key=config_type_dict_by_key,
    )


def create_environment_type(pipeline_def, mode=None):
    return create_environment_schema(pipeline_def, mode).environment_type



from dagster import (
    DependencyDefinition,
    InProcessExecutorConfig,
    InputDefinition,
    MultiprocessExecutorConfig,
    PipelineDefinition,
    RunConfig,
    execute_pipeline,
    lambda_solid,
)

from dagster.core.runs import RunStorageMode


def test_diamond_simple_execution():
    result = execute_pipeline(
        define_diamond_pipeline(), run_config=RunConfig(executor_config=InProcessExecutorConfig())
    )
    assert result.success
    assert result.result_for_solid('adder').transformed_value() == 11


def transform_event(result, solid_name):
    return result.result_for_solid(solid_name).transforms[0]


def test_diamond_multi_execution():
    pipeline = define_diamond_pipeline()
    result = execute_pipeline(
        pipeline,
        run_config=RunConfig(
            executor_config=MultiprocessExecutorConfig(define_diamond_pipeline),
            storage_mode=RunStorageMode.FILESYSTEM,
        ),
    )
    assert result.success

    # FIXME: be able to get this value
    # https://github.com/dagster-io/dagster/issues/953
    # assert result.result_for_solid('adder').transformed_value() == 11

    pids_by_solid = {}
    for solid in pipeline.solids:
        pids_by_solid[solid.name] = transform_event(result, solid.name).logging_tags['pid']

    # guarantee that all solids ran in their own process
    assert len(set(pids_by_solid.values())) == len(pipeline.solids)


def define_diamond_pipeline():
    @lambda_solid
    def return_two():
        return 2

    @lambda_solid(inputs=[InputDefinition('num')])
    def add_three(num):
        return num + 3

    @lambda_solid(inputs=[InputDefinition('num')])
    def mult_three(num):
        return num * 3

    @lambda_solid(inputs=[InputDefinition('left'), InputDefinition('right')])
    def adder(left, right):
        return left + right

    return PipelineDefinition(
        name='diamond_execution',
        solids=[return_two, add_three, mult_three, adder],
        dependencies={
            'add_three': {'num': DependencyDefinition('return_two')},
            'mult_three': {'num': DependencyDefinition('return_two')},
            'adder': {
                'left': DependencyDefinition('add_three'),
                'right': DependencyDefinition('mult_three'),
            },
        },
    )


def define_error_pipeline():
    @lambda_solid
    def throw_error():
        raise Exception('bad programmer')

    return PipelineDefinition(name='error_pipeline', solids=[throw_error])


def test_error_pipeline():
    pipeline = define_error_pipeline()
    result = execute_pipeline(
        pipeline,
        run_config=RunConfig(executor_config=InProcessExecutorConfig(raise_on_error=False)),
    )
    assert not result.success


def test_error_pipeline_multiprocess():
    pipeline = define_error_pipeline()
    result = execute_pipeline(
        pipeline,
        run_config=RunConfig(
            executor_config=MultiprocessExecutorConfig(define_error_pipeline),
            storage_mode=RunStorageMode.FILESYSTEM,
        ),
    )
    assert not result.success



from dagster import check

from dagster.core.types import Field, Dict
from dagster.core.types.field_utils import check_user_facing_opt_field_param
from dagster.core.execution.user_context import ExecutionContext
from dagster.core.system_config.objects import DEFAULT_CONTEXT_NAME

from dagster.utils.logging import LogLevelEnum, level_from_string

from .resource import ResourceDefinition


class PipelineContextDefinition(object):
    '''Defines a context type supported by a pipeline.

    Pipelines declare the different context types they support, in the form
    of PipelineContextDefinitions. For example a pipeline could declare a context
    definition for different operating environments: unittest, integration tests,
    production and so forth. The user provides a context function that returns an
    ``ExecutionContext`` that is passed to every solid. One can hang resources
    (such as db connections) off of that context. Thus the pipeline author
    has complete control over how the author of each individual solid within
    the pipeline interacts with its operating environment.

    The ``PipelineContextDefinition`` is passed to the ``PipelineDefinition`` in
    a dictionary keyed by its name so the name is not present in this object.

    Args:
        context_fn (Callable):
            Signature of context_fn:
            (pipeline: PipelineDefintion, config_value: Any) => ExecutionContext

            Returns *or* yields an ExecutionContext.

            If it yields a context, the code after the yield executes after pipeline
            completion, just like a python context manager.

        config_field (Field):
            Define the configuration for the context

        description (str)

    Attributes:
        config_field (Field): The configuration for the pipeline context.

        context_fn (callable):
            Signature is (**pipeline**: `PipelineDefintion`, **config_value**: `Any`) :
            `ExecutionContext`.

            A callable that either returns *or* yields an ``ExecutionContext``.

        description (str): A description of what this context represents
    '''

    @staticmethod
    def passthrough_context_definition(context_params):
        '''Create a context definition from a pre-existing context. This can be useful
        in testing contexts where you may want to create a context manually and then
        pass it into a one-off PipelineDefinition

        Args:
            context (ExecutionContext): The context that will provided to the pipeline.
        Returns:
            PipelineContextDefinition: The passthrough context definition.
        '''

        check.inst_param(context_params, 'context', ExecutionContext)
        context_definition = PipelineContextDefinition(context_fn=lambda *_args: context_params)
        return {DEFAULT_CONTEXT_NAME: context_definition}

    def __init__(self, context_fn=None, config_field=None, resources=None, description=None):
        if config_field is None and context_fn is None:
            config_field = _default_config_field()
            context_fn = _default_context_fn

        self.config_field = check_user_facing_opt_field_param(
            config_field, 'config_field', 'of a PipelineContextDefinition'
        )
        self.context_fn = check.opt_callable_param(
            context_fn, 'context_fn', lambda *args, **kwargs: ExecutionContext()
        )
        self.resource_defs = check.opt_dict_param(
            resources, 'resources', key_type=str, value_type=ResourceDefinition
        )
        self.description = description


def _default_context_fn(init_context):
    log_level = level_from_string(init_context.context_config['log_level'])
    return ExecutionContext.console_logging(log_level)


def _default_config_field():
    return Field(Dict({'log_level': Field(LogLevelEnum, is_optional=True, default_value='INFO')}))


def default_pipeline_context_definitions():
    return {
        DEFAULT_CONTEXT_NAME: PipelineContextDefinition(
            config_field=_default_config_field(), context_fn=_default_context_fn
        )
    }



import six

from dagster import check
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.utils import toposort_flatten
from dagster.core.errors import DagsterInvalidDefinitionError

from .context import PipelineContextDefinition, default_pipeline_context_definitions
from .dependency import DependencyDefinition, DependencyStructure, Solid, SolidHandle, SolidInstance
from .mode import ModeDefinition
from .pipeline_creation import create_execution_structure, construct_runtime_type_dictionary
from .solid import ISolidDefinition


def _check_solids_arg(pipeline_name, solid_defs):
    if not isinstance(solid_defs, list):
        raise DagsterInvalidDefinitionError(
            '"solids" arg to pipeline "{name}" is not a list. Got {val}.'.format(
                name=pipeline_name, val=repr(solid_defs)
            )
        )
    for solid_def in solid_defs:
        if isinstance(solid_def, ISolidDefinition):
            continue
        elif callable(solid_def):
            raise DagsterInvalidDefinitionError(
                '''You have passed a lambda or function {func} into pipeline {name} that is
                not a solid. You have likely forgetten to annotate this function with
                an @solid or @lambda_solid decorator.'
                '''.format(
                    name=pipeline_name, func=solid_def.__name__
                )
            )
        else:
            raise DagsterInvalidDefinitionError(
                'Invalid item in solid list: {item}'.format(item=repr(solid_def))
            )

    return solid_defs


class PipelineDefinition(object):
    '''A instance of a PipelineDefinition represents a pipeline in dagster.

    A pipeline is comprised of:

    - Solids:
        Each solid represents a functional unit of data computation.

    - Context Definitions:
        Pipelines can be designed to execute in a number of different operating environments
        (e.g. prod, dev, unittest) that require different configuration and setup. A context
        definition defines how a context (of type ExecutionContext) is created and what
        configuration is necessary to create it.

    - Dependencies:
        Solids within a pipeline are arranged as a DAG (directed, acyclic graph). Dependencies
        determine how the values produced by solids flow through the DAG.

    Args:
        solids (List[SolidDefinition]):
            The set of solid definitions used in this pipeline.
        name (Optional[str])
        despcription (Optional[str])
        context_definitions (Optional[Dict[str, PipelineContextDefinition]]):
            A mapping of context names to PipelineContextDefinition.
        dependencies (Optional[Dict[Union[str, SolidInstance], Dict[str, DependencyDefinition]]]):
            A structure that declares where each solid gets its inputs. The keys at the top
            level dict are either string names of solids or SolidInstances. The values
            are dicts that map input names to DependencyDefinitions.


    Attributes:
        name (str):
            Name of the pipeline. Must be unique per-repository.
        description (str):
            Description of the pipeline. Optional.
        solids (List[SolidDefinition]):
            List of the solids in this pipeline.
        dependencies (Dict[str, Dict[str, DependencyDefinition]]) :
            Dependencies that constitute the structure of the pipeline. This is a two-dimensional
            array that maps solid_name => input_name => DependencyDefinition instance
        context_definitions (Dict[str, PipelineContextDefinition]):
            The context definitions available for consumers of this pipelines. For example, a
            unit-testing environment and a production environment probably have very different
            configuration and requirements. There would be one context definition per
            environment.

            Only one context will be used at runtime, selected by environment configuration.
        dependency_structure (DependencyStructure):
            Used mostly internally. This has the same information as the dependencies data
            structure, but indexed for fast usage.
    '''

    def __init__(
        self,
        solids,
        name=None,
        description=None,
        context_definitions=None,
        dependencies=None,
        mode_definitions=None,
    ):
        self.name = check.opt_str_param(name, 'name', '<<unnamed>>')
        self.description = check.opt_str_param(description, 'description')

        check.invariant(
            not (context_definitions and mode_definitions),
            'Cannot specify both context_definitions and modes',
        )

        self.mode_definitions = check.opt_list_param(
            mode_definitions, 'mode_definitions', of_type=ModeDefinition
        )

        solids = check.list_param(
            _check_solids_arg(self.name, solids), 'solids', of_type=ISolidDefinition
        )

        if context_definitions is None:
            context_definitions = default_pipeline_context_definitions()

        self.context_definitions = check.dict_param(
            context_definitions,
            'context_definitions',
            key_type=str,
            value_type=PipelineContextDefinition,
        )

        self.dependencies = _validate_dependency_dict(dependencies)

        dependency_structure, pipeline_solid_dict = create_execution_structure(
            solids, self.dependencies
        )

        self._solid_dict = pipeline_solid_dict
        self.dependency_structure = dependency_structure

        self._runtime_type_dict = construct_runtime_type_dictionary(solids)

        # Validate solid resource dependencies
        _validate_resource_dependencies(self.context_definitions, solids)

    def _get_mode_definition(self, mode):
        check.str_param(mode, 'mode')
        for mode_definition in self.mode_definitions:
            if mode_definition.name == mode:
                return mode_definition

        return None

    @property
    def is_modeless(self):
        return not self.mode_definitions

    @property
    def is_single_mode(self):
        return len(self.mode_definitions) == 1

    def has_mode_definition(self, mode):
        check.str_param(mode, 'mode')
        return bool(self._get_mode_definition(mode))

    def get_mode_definition(self, mode=None):
        check.opt_str_param(mode, 'mode')
        if mode is None:
            check.invariant(self.is_modeless or self.is_single_mode)
            if self.is_modeless:
                return None
            else:
                check.invariant(self.is_single_mode)
                return self.mode_definitions[0]

        mode_def = self._get_mode_definition(mode)

        if mode_def is None:
            check.failed(
                'Could not find mode {mode} in pipeline {name}'.format(mode=mode, name=self.name)
            )

        return mode_def

    @property
    def available_modes(self):
        return [mode_def.name for mode_def in self.mode_definitions]

    @property
    def display_name(self):
        '''Name suitable for exception messages, logging etc. If pipeline
        is unnamed the method with return "<<unnamed>>".

        Returns:
            str: Display name of pipeline
        '''
        return self.name if self.name else '<<unnamed>>'

    @property
    def solids(self):
        '''Return the solids in the pipeline.

        Returns:
            List[Solid]: List of solids.
        '''
        return list(set(self._solid_dict.values()))

    def has_solid_named(self, name):
        '''Return whether or not the solid is in the piepline

        Args:
            name (str): Name of solid

        Returns:
            bool: True if the solid is in the pipeline
        '''
        check.str_param(name, 'name')
        return name in self._solid_dict

    def solid_named(self, name):
        '''Return the solid named "name". Throws if it does not exist.

        Args:
            name (str): Name of solid

        Returns:
            SolidDefinition: SolidDefinition with correct name.
        '''
        check.str_param(name, 'name')
        if name not in self._solid_dict:
            raise DagsterInvariantViolationError(
                'Pipeline {pipeline_name} has no solid named {name}.'.format(
                    pipeline_name=self.name, name=name
                )
            )
        return self._solid_dict[name]

    def get_solid(self, handle):
        check.inst_param(handle, 'handle', SolidHandle)
        current = handle
        lineage = []
        while current:
            lineage.append(current.name)
            current = current.parent

        name = lineage.pop()
        solid = self.solid_named(name)
        while lineage:
            name = lineage.pop()
            solid = solid.definition.solid_named(name)

        return solid

    def has_runtime_type(self, name):
        check.str_param(name, 'name')
        return name in self._runtime_type_dict

    def runtime_type_named(self, name):
        check.str_param(name, 'name')
        return self._runtime_type_dict[name]

    def has_context(self, name):
        check.str_param(name, 'name')
        return name in self.context_definitions

    def all_runtime_types(self):
        return self._runtime_type_dict.values()

    @property
    def solid_defs(self):
        return list({solid.definition for solid in self.solids})

    def solid_def_named(self, name):
        check.str_param(name, 'name')

        for solid in self.solids:
            if solid.definition.name == name:
                return solid.definition

        check.failed('{} not found'.format(name))

    def has_solid_def(self, name):
        check.str_param(name, 'name')

        for solid in self.solids:
            if solid.definition.name == name:
                return True

        return False

    def build_sub_pipeline(self, solid_subset):
        return self if solid_subset is None else _build_sub_pipeline(self, solid_subset)


def _create_adjacency_lists(solids, dep_structure):
    check.list_param(solids, 'solids', Solid)
    check.inst_param(dep_structure, 'dep_structure', DependencyStructure)

    visit_dict = {s.name: False for s in solids}
    forward_edges = {s.name: set() for s in solids}
    backward_edges = {s.name: set() for s in solids}

    def visit(solid_name):
        if visit_dict[solid_name]:
            return

        visit_dict[solid_name] = True

        for output_handle in dep_structure.deps_of_solid(solid_name):
            forward_node = output_handle.solid.name
            backward_node = solid_name
            if forward_node in forward_edges:
                forward_edges[forward_node].add(backward_node)
                backward_edges[backward_node].add(forward_node)
                visit(forward_node)

    for s in solids:
        visit(s.name)

    return (forward_edges, backward_edges)


def solids_in_topological_order(pipeline):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    _forward_edges, backward_edges = _create_adjacency_lists(
        pipeline.solids, pipeline.dependency_structure
    )

    order = toposort_flatten(backward_edges)
    return [pipeline.solid_named(solid_name) for solid_name in order]


def _dep_key_of(solid):
    return SolidInstance(solid.definition.name, solid.name)


def _build_sub_pipeline(pipeline_def, solid_names):
    '''
    Build a pipeline which is a subset of another pipeline.
    Only includes the solids which are in solid_names.
    '''

    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.list_param(solid_names, 'solid_names', of_type=str)

    solid_name_set = set(solid_names)
    solids = list(map(pipeline_def.solid_named, solid_names))
    deps = {_dep_key_of(solid): {} for solid in solids}

    def _out_handle_of_inp(input_handle):
        if pipeline_def.dependency_structure.has_dep(input_handle):
            output_handle = pipeline_def.dependency_structure.get_dep(input_handle)
            if output_handle.solid.name in solid_name_set:
                return output_handle
        return None

    for solid in solids:
        for input_handle in solid.input_handles():
            output_handle = _out_handle_of_inp(input_handle)
            if output_handle:
                deps[_dep_key_of(solid)][input_handle.input_def.name] = DependencyDefinition(
                    solid=output_handle.solid.name, output=output_handle.output_def.name
                )

    return PipelineDefinition(
        name=pipeline_def.name,
        solids=list({solid.definition for solid in solids}),
        context_definitions=pipeline_def.context_definitions,
        dependencies=deps,
    )


def _validate_dependency_dict(dependencies):
    prelude = (
        'The expected type for "dependencies" is dict[Union[str, SolidInstance], dict[str, '
        'DependencyDefinition]]. '
    )

    if dependencies is None:
        return {}

    if not isinstance(dependencies, dict):
        raise DagsterInvalidDefinitionError(
            prelude
            + 'Received value {val} of type {type} at the top level.'.format(
                val=dependencies, type=type(dependencies)
            )
        )

    for key, dep_dict in dependencies.items():
        if not (isinstance(key, six.string_types) or isinstance(key, SolidInstance)):
            raise DagsterInvalidDefinitionError(
                prelude + 'Expected str or SolidInstance key in the top level dict. '
                'Received value {val} of type {type}'.format(val=key, type=type(key))
            )
        if not isinstance(dep_dict, dict):
            if isinstance(dep_dict, DependencyDefinition):
                raise DagsterInvalidDefinitionError(
                    prelude + 'Received a DependencyDefinition one layer too high under key {key}. '
                    'The DependencyDefinition should be moved in to a dict keyed on '
                    'input name.'.format(key=key)
                )
            else:
                raise DagsterInvalidDefinitionError(
                    prelude + 'Under key {key} received value {val} of type {type}. '
                    'Expected dict[str, DependencyDefinition]'.format(
                        key=key, val=dep_dict, type=type(dep_dict)
                    )
                )

        for input_key, dep in dep_dict.items():
            if not isinstance(input_key, six.string_types):
                raise DagsterInvalidDefinitionError(
                    prelude
                    + 'Received non-sting key in the inner dict for key {key}.'.format(key=key)
                )
            if not isinstance(dep, DependencyDefinition):
                raise DagsterInvalidDefinitionError(
                    prelude
                    + 'Expected DependencyDefinition for solid "{key}" input "{input_key}". '
                    'Received value {val} of type {type}.'.format(
                        key=key, input_key=input_key, val=dep, type=type(dep)
                    )
                )

    return dependencies


def _validate_resource_dependencies(context_definitions, solids):
    '''This validation ensures that each pipeline context provides the resources that are required
    by each solid.
    '''
    check.dict_param(
        context_definitions, 'context_definitions', str, value_type=PipelineContextDefinition
    )
    check.list_param(solids, 'solids', of_type=ISolidDefinition)

    for context_name, context in context_definitions.items():
        context_resources = set(context.resource_defs.keys())

        for solid in solids:
            for resource in solid.resources:
                if resource not in context_resources:
                    raise DagsterInvalidDefinitionError(
                        (
                            'Resource "{resource}" is required by solid {solid_name}, but is not '
                            'provided by context "{context_name}"'
                        ).format(
                            resource=resource, solid_name=solid.name, context_name=context_name
                        )
                    )



from dagster import check

from .resource import ResourceDefinition


class ModeDefinition:
    def __init__(self, name='default', resources=None):
        self.name = check.str_param(name, 'name')
        self.resource_defs = check.opt_dict_param(
            resources, 'resources', key_type=str, value_type=ResourceDefinition
        )



from collections import namedtuple

import warnings

from dagster import check

from dagster.core.definitions.pipeline import PipelineDefinition
from dagster.core.definitions.resource import ResourceDefinition


class InitContext(namedtuple('_InitContext', 'context_config pipeline_def run_id')):
    '''
    InitContext is the context object provided to context creation functions.
    In effect, it is the state available to those functions, and any function that
    is called prior to pipeline execution, plus the configuration value for that context.

    Attributes:
        context_config (Any):
            The configuration data provided by the environment config. The schema for this 
            data is defined by ``config_field`` on the :py:class:`PipelineContextDefinition`
        pipeline_def (PipelineDefinition): The pipeline definition currently being executed.
        run_id (str): The ID for this run of the pipeline.

    '''

    def __new__(cls, context_config, pipeline_def, run_id):
        return super(InitContext, cls).__new__(
            cls,
            context_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.str_param(run_id, 'run_id'),
        )

    @property
    def config(self):
        warnings.warn(
            (
                'As of 0.3.2 the config property is deprecated. Use context_config instead. '
                'This will be removed in 0.4.0.'
            )
        )
        return self.context_config


class InitResourceContext(
    namedtuple(
        'InitResourceContext', 'context_config resource_config pipeline_def resource_def run_id'
    )
):
    '''
    Similar to InitContext, but is resource specific. It includes all the properties
    in the InitContext, plus the resource config and the resource definition.


    Attributes:
        context_config (Any):
            The configuration data provided by the environment config. The schema for this
            data is defined by ``config_field`` on the :py:class:`PipelineContextDefinition`
        resource_config (Any): 
            The configuration data provided by the environment config. The schema for this
            data is defined by ``config_field`` on the :py:class:`ResourceDefinition`
        pipeline_def (PipelineDefinition): The pipeline definition currently being executed.
        resource_def (ResourceDefinition): The resource definition for the resource being constructed.
        run_id (str): The ID for this run of the pipeline.
    '''

    def __new__(cls, context_config, resource_config, pipeline_def, resource_def, run_id):
        return super(InitResourceContext, cls).__new__(
            cls,
            context_config,
            resource_config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.inst_param(resource_def, 'resource_def', ResourceDefinition),
            check.str_param(run_id, 'run_id'),
        )



from contextlib import contextmanager
import inspect
import sys
import time

from contextlib2 import ExitStack

from dagster import check
from dagster.utils import merge_dicts
from dagster.utils.logging import define_colored_console_logger
from dagster.utils.error import serializable_error_info_from_exc_info

from dagster.core.definitions import PipelineDefinition, create_environment_type
from dagster.core.definitions.resource import ResourcesBuilder, ResourcesSource
from dagster.core.definitions.environment_configs import construct_environment_config


from dagster.core.errors import (
    DagsterError,
    DagsterInvariantViolationError,
    DagsterUserCodeExecutionError,
    DagsterContextFunctionError,
    DagsterResourceFunctionError,
    user_code_error_boundary,
)

from dagster.core.events import DagsterEvent, PipelineInitFailureData
from dagster.core.events.logging import construct_event_logger


from dagster.core.intermediates_manager import (
    ObjectStoreIntermediatesManager,
    InMemoryIntermediatesManager,
    IntermediatesManager,
)

from dagster.core.log import DagsterLog

from dagster.core.object_store import FileSystemObjectStore, construct_type_storage_plugin_registry

from dagster.core.runs import (
    DagsterRunMeta,
    FileSystemRunStorage,
    InMemoryRunStorage,
    RunStorage,
    RunStorageMode,
)

from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.types.evaluator import (
    EvaluationError,
    evaluate_config_value,
    friendly_string_for_error,
)

from .execution_context import (
    RunConfig,
    SystemPipelineExecutionContextData,
    SystemPipelineExecutionContext,
)
from .init_context import InitContext, InitResourceContext
from .resource_creation_adapter import ResourceCreationAdapter
from .user_context import ExecutionContext


class PipelineConfigEvaluationError(Exception):
    def __init__(self, pipeline, errors, config_value, *args, **kwargs):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.errors = check.list_param(errors, 'errors', of_type=EvaluationError)
        self.config_value = config_value

        error_msg = 'Pipeline "{pipeline}" config errors:'.format(pipeline=pipeline.name)

        error_messages = []

        for i_error, error in enumerate(self.errors):
            error_message = friendly_string_for_error(error)
            error_messages.append(error_message)
            error_msg += '\n    Error {i_error}: {error_message}'.format(
                i_error=i_error + 1, error_message=error_message
            )

        self.message = error_msg
        self.error_messages = error_messages

        super(PipelineConfigEvaluationError, self).__init__(error_msg, *args, **kwargs)


def create_environment_config(pipeline, environment_dict=None, mode=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(environment_dict, 'environment')
    check.opt_str_param(mode, 'mode')

    environment_type = create_environment_type(pipeline, mode)

    result = evaluate_config_value(environment_type, environment_dict)

    if not result.success:
        raise PipelineConfigEvaluationError(pipeline, result.errors, environment_dict)

    return construct_environment_config(result.value)


def construct_run_storage(run_config, environment_config):
    '''
    Construct the run storage for this pipeline. Our rules are the following:

    If the RunConfig has a storage_mode provided, we use that.

    Then we fallback to environment config.

    If there is no config, we default to in memory storage. This is mostly so
    that tests default to in-memory.
    '''
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)

    if run_config.storage_mode:
        if run_config.storage_mode == RunStorageMode.FILESYSTEM:
            return FileSystemRunStorage()
        elif run_config.storage_mode == RunStorageMode.IN_MEMORY:
            return InMemoryRunStorage()
        elif run_config.storage_mode == RunStorageMode.S3:
            # TODO: Revisit whether we want to use S3 run storage
            return FileSystemRunStorage()
        else:
            check.failed('Unexpected enum {}'.format(run_config.storage_mode))
    elif environment_config.storage.storage_mode == 'filesystem':
        return FileSystemRunStorage()
    elif environment_config.storage.storage_mode == 'in_memory':
        return InMemoryRunStorage()
    elif environment_config.storage.storage_mode == 's3':
        # TODO: Revisit whether we want to use S3 run storage
        return FileSystemRunStorage()
    elif environment_config.storage.storage_mode is None:
        return InMemoryRunStorage()
    else:
        raise DagsterInvariantViolationError(
            'Invalid storage specified {}'.format(environment_config.storage.storage_mode)
        )


@contextmanager
def scoped_pipeline_context(pipeline_def, environment_config, run_config, intermediates_manager):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    run_storage = construct_run_storage(run_config, environment_config)

    run_storage.write_dagster_run_meta(
        DagsterRunMeta(
            run_id=run_config.run_id, timestamp=time.time(), pipeline_name=pipeline_def.name
        )
    )

    init_context = InitContext(
        context_config=environment_config.context.config if environment_config.context else {},
        pipeline_def=pipeline_def,
        run_id=run_config.run_id,
    )

    try:

        context_definition = (
            pipeline_def.context_definitions[environment_config.context.name]
            if environment_config.context
            else None
        )

        import logging

        with user_code_context_manager(
            lambda: (
                context_definition.context_fn(init_context)
                if context_definition
                # hardcoded default for now until we have top-level logging as well
                else ExecutionContext.console_logging(logging.INFO)
            ),
            DagsterContextFunctionError,
            'Error executing context_fn on ContextDefinition {name}'.format(
                name=environment_config.context.name if environment_config.context else 'NOCONTEXT'
            ),
        ) as execution_context:
            check.inst(execution_context, ExecutionContext)

            resource_creation_adapter = ResourceCreationAdapter(
                execution_context=execution_context,
                context_definition=context_definition,
                mode_definition=pipeline_def.get_mode_definition(run_config.mode),
                environment_config=environment_config,
            )

            with _create_resources(
                pipeline_def, resource_creation_adapter, environment_config, run_config.run_id
            ) as resources:

                yield construct_pipeline_execution_context(
                    run_config,
                    execution_context,
                    pipeline_def,
                    resources,
                    environment_config,
                    run_storage,
                    intermediates_manager,
                )

    except DagsterError as dagster_error:
        user_facing_exc_info = (
            # pylint does not know original_exc_info exists is is_user_code_error is true
            # pylint: disable=no-member
            dagster_error.original_exc_info
            if dagster_error.is_user_code_error
            else sys.exc_info()
        )

        if run_config.executor_config.raise_on_error:
            raise dagster_error

        error_info = serializable_error_info_from_exc_info(user_facing_exc_info)
        yield DagsterEvent.pipeline_init_failure(
            pipeline_name=pipeline_def.name,
            failure_data=PipelineInitFailureData(error=error_info),
            log=_create_context_free_log(run_config, pipeline_def),
        )


@contextmanager
def yield_pipeline_execution_context(pipeline_def, environment_dict, run_config):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(environment_dict, 'environment_dict', key_type=str)
    check.inst_param(run_config, 'run_config', RunConfig)

    environment_config = create_environment_config(pipeline_def, environment_dict)
    intermediates_manager = construct_intermediates_manager(
        run_config, environment_config, pipeline_def
    )
    with scoped_pipeline_context(
        pipeline_def, environment_config, run_config, intermediates_manager
    ) as context:
        yield context


def construct_pipeline_execution_context(
    run_config,
    execution_context,
    pipeline,
    resources,
    environment_config,
    run_storage,
    intermediates_manager,
):
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(execution_context, 'execution_context', ExecutionContext)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_inst_param(resources, 'resources', ResourcesBuilder)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(run_storage, 'run_storage', RunStorage)
    check.inst_param(intermediates_manager, 'intermediates_manager', IntermediatesManager)

    loggers = _create_loggers(run_config, execution_context)
    logging_tags = get_logging_tags(execution_context, run_config, pipeline)
    log = DagsterLog(run_config.run_id, logging_tags, loggers)

    return SystemPipelineExecutionContext(
        SystemPipelineExecutionContextData(
            pipeline_def=pipeline,
            run_config=run_config,
            resources=resources,
            environment_config=environment_config,
            run_storage=run_storage,
            intermediates_manager=intermediates_manager,
        ),
        logging_tags=logging_tags,
        log=log,
    )


@contextmanager
def _create_resources(pipeline_def, resource_creation_adapter, environment, run_id):
    check.inst_param(
        resource_creation_adapter, 'resource_creation_adapter', ResourceCreationAdapter
    )

    if resource_creation_adapter.is_resource_override:
        yield ResourcesBuilder(
            resource_creation_adapter.get_override_resources(),
            ResourcesSource.CUSTOM_EXECUTION_CONTEXT,
        )
        return

    resources = {}

    # See https://bit.ly/2zIXyqw
    # The "ExitStack" allows one to stack up N context managers and then yield
    # something. We do this so that resources can cleanup after themselves. We
    # can potentially have many resources so we need to use this abstraction.
    with ExitStack() as stack:
        for resource_name, resource_def in resource_creation_adapter.resource_defs.items():
            user_fn = _create_resource_fn_lambda(
                pipeline_def,
                resource_def,
                resource_creation_adapter.get_resource_config(resource_name),
                environment,
                run_id,
            )

            resource_obj = stack.enter_context(
                user_code_context_manager(
                    user_fn,
                    DagsterResourceFunctionError,
                    'Error executing resource_fn on ResourceDefinition {name}'.format(
                        name=resource_name
                    ),
                )
            )

            resources[resource_name] = resource_obj
        yield ResourcesBuilder(resources, ResourcesSource.PIPELINE_CONTEXT_DEF)


def _create_resource_fn_lambda(pipeline_def, resource_def, resource_config, environment, run_id):
    return lambda: resource_def.resource_fn(
        InitResourceContext(
            pipeline_def=pipeline_def,
            resource_def=resource_def,
            context_config=environment.context.config if environment.context else None,
            resource_config=resource_config,
            run_id=run_id,
        )
    )


def ensure_dagster_aws_requirements():
    try:
        import dagster_aws
    except (ImportError, ModuleNotFoundError):
        raise check.CheckError(
            'dagster_aws must be available for import in order to make use of an S3ObjectStore'
        )

    return dagster_aws


def construct_intermediates_manager(run_config, environment_config, pipeline_def):
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

    if run_config.storage_mode:
        if run_config.storage_mode == RunStorageMode.FILESYSTEM:
            return ObjectStoreIntermediatesManager(
                FileSystemObjectStore(
                    run_config.run_id,
                    construct_type_storage_plugin_registry(pipeline_def, RunStorageMode.FILESYSTEM),
                )
            )
        elif run_config.storage_mode == RunStorageMode.IN_MEMORY:
            return InMemoryIntermediatesManager()
        elif run_config.storage_mode == RunStorageMode.S3:
            _dagster_aws = ensure_dagster_aws_requirements()
            from dagster_aws.s3.object_store import S3ObjectStore

            return ObjectStoreIntermediatesManager(
                S3ObjectStore(
                    environment_config.storage.storage_config['s3_bucket'],
                    run_config.run_id,
                    construct_type_storage_plugin_registry(pipeline_def, RunStorageMode.S3),
                )
            )
        else:
            check.failed('Unexpected enum {}'.format(run_config.storage_mode))
    elif environment_config.storage.storage_mode == 'filesystem':
        return ObjectStoreIntermediatesManager(
            FileSystemObjectStore(
                run_config.run_id,
                construct_type_storage_plugin_registry(pipeline_def, RunStorageMode.FILESYSTEM),
            )
        )
    elif environment_config.storage.storage_mode == 'in_memory':
        return InMemoryIntermediatesManager()
    elif environment_config.storage.storage_mode == 's3':
        _dagster_aws = ensure_dagster_aws_requirements()
        from dagster_aws.s3.object_store import S3ObjectStore

        return ObjectStoreIntermediatesManager(
            S3ObjectStore(
                environment_config.storage.storage_config['s3_bucket'],
                run_config.run_id,
                construct_type_storage_plugin_registry(pipeline_def, RunStorageMode.S3),
            )
        )
    elif environment_config.storage.storage_mode is None:
        return InMemoryIntermediatesManager()
    else:
        raise DagsterInvariantViolationError(
            'Invalid storage specified {}'.format(environment_config.storage.storage_mode)
        )


def _create_loggers(run_config, execution_context):
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(execution_context, 'execution_context', ExecutionContext)

    if run_config.event_callback:
        return execution_context.loggers + [construct_event_logger(run_config.event_callback)]
    elif run_config.loggers:
        return execution_context.loggers + run_config.loggers
    else:
        return execution_context.loggers


def _create_context_free_log(run_config, pipeline_def):
    '''In the event of pipeline initialization failure, we want to be able to log the failure
    without a dependency on the ExecutionContext to initialize DagsterLog
    '''
    check.inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

    # Use the default logger
    loggers = [define_colored_console_logger('dagster')]
    if run_config.event_callback:
        loggers += [construct_event_logger(run_config.event_callback)]
    elif run_config.loggers:
        loggers += run_config.loggers

    return DagsterLog(run_config.run_id, get_logging_tags(None, run_config, pipeline_def), loggers)


def _ensure_gen(thing_or_gen):
    if not inspect.isgenerator(thing_or_gen):

        def _gen_thing():
            yield thing_or_gen

        return _gen_thing()

    return thing_or_gen


@contextmanager
def user_code_context_manager(user_fn, error_cls, msg):
    '''Wraps the output of a user provided function that may yield or return a value and
    returns a generator that asserts it only yields a single value.
    '''
    check.callable_param(user_fn, 'user_fn')
    check.subclass_param(error_cls, 'error_cls', DagsterUserCodeExecutionError)

    with user_code_error_boundary(error_cls, msg):
        thing_or_gen = user_fn()
        gen = _ensure_gen(thing_or_gen)

        try:
            thing = next(gen)
        except StopIteration:
            check.failed('Must yield one item. You did not yield anything.')

        yield thing

        stopped = False

        try:
            next(gen)
        except StopIteration:
            stopped = True

        check.invariant(stopped, 'Must yield one item. Yielded more than one item')


def get_logging_tags(user_context_params, run_config, pipeline):
    check.opt_inst_param(user_context_params, 'user_context_params', ExecutionContext)
    check.opt_inst_param(run_config, 'run_config', RunConfig)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    user_tags = user_context_params.tags if user_context_params else {}
    run_tags = run_config.tags if run_config else {}

    base_tags = merge_dicts({'pipeline': pipeline.name}, user_tags)

    if run_config and run_config.tags:
        user_keys = set(user_tags.keys())
        provided_keys = set(run_tags.keys())
        if not user_keys.isdisjoint(provided_keys):
            raise DagsterInvariantViolationError(
                (
                    'You have specified tags and user-defined tags '
                    'that overlap. User keys: {user_keys}. Reentrant keys: '
                    '{provided_keys}.'
                ).format(user_keys=user_keys, provided_keys=provided_keys)
            )

        return merge_dicts(base_tags, run_tags)
    else:
        return base_tags



import json

from dagster import check
from dagster.utils.error import SerializableErrorInfo

from dagster.utils.logging import (
    DEBUG,
    StructuredLoggerHandler,
    JsonEventLoggerHandler,
    StructuredLoggerMessage,
    check_valid_level_param,
    construct_single_handler_logger,
)


class EventRecord(object):
    def __init__(
        self,
        error_info,
        message,
        level,
        user_message,
        run_id,
        timestamp,
        step_key=None,
        pipeline_name=None,
        dagster_event=None,
    ):
        from dagster.core.events import DagsterEvent

        self._error_info = check.opt_inst_param(error_info, 'error_info', SerializableErrorInfo)
        self._message = check.str_param(message, 'message')
        self._level = check_valid_level_param(level)
        self._user_message = check.str_param(user_message, 'user_message')
        self._run_id = check.str_param(run_id, 'run_id')
        self._timestamp = check.float_param(timestamp, 'timestamp')
        self._step_key = check.opt_str_param(step_key, 'step_key')
        self._pipeline_name = check.opt_str_param(pipeline_name, 'pipeline_name')
        self._dagster_event = check.opt_inst_param(dagster_event, 'dagster_event', DagsterEvent)

    @property
    def message(self):
        return self._message

    @property
    def level(self):
        return self._level

    @property
    def user_message(self):
        return self._user_message

    @property
    def is_dagster_event(self):
        return bool(self._dagster_event)

    @property
    def run_id(self):
        return self._run_id

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def error_info(self):
        return self._error_info

    @property
    def step_key(self):
        return self._step_key

    @property
    def pipeline_name(self):
        return self._pipeline_name

    @property
    def dagster_event(self):
        return self._dagster_event

    def to_dict(self):
        return {
            'run_id': self.run_id,
            'message': self.message,
            'level': self.level,
            'user_message': self.user_message,
            'timestamp': self.timestamp,
            'error_info': self.error_info,
            'step_key': self.step_key,
            'dagster_event': self.dagster_event,
            'pipeline_name': self.pipeline_name,
        }


class DagsterEventRecord(EventRecord):
    pass


class LogMessageRecord(EventRecord):
    pass


def construct_error_info(logger_message):
    if 'error_info' not in logger_message.meta:
        return None

    raw_error_info = logger_message.meta['error_info']

    message, stack = json.loads(raw_error_info)

    return SerializableErrorInfo(message, stack)


def construct_event_record(logger_message):
    check.inst_param(logger_message, 'logger_message', StructuredLoggerMessage)

    log_record_cls = LogMessageRecord
    if logger_message.meta.get('dagster_event'):
        log_record_cls = DagsterEventRecord

    return log_record_cls(
        message=logger_message.message,
        level=logger_message.level,
        user_message=logger_message.meta['orig_message'],
        run_id=logger_message.meta['run_id'],
        timestamp=logger_message.record.created,
        step_key=logger_message.meta.get('step_key'),
        pipeline_name=logger_message.meta.get('pipeline_name'),
        dagster_event=logger_message.meta.get('dagster_event'),
        error_info=construct_error_info(logger_message),
    )


def construct_event_logger(event_record_callback):
    '''
    Callback receives a stream of event_records
    '''
    check.callable_param(event_record_callback, 'event_record_callback')

    return construct_single_handler_logger(
        'event-logger',
        DEBUG,
        StructuredLoggerHandler(
            lambda logger_message: event_record_callback(construct_event_record(logger_message))
        ),
    )


def construct_json_event_logger(json_path):
    '''Record a stream of event records to json'''
    check.str_param(json_path, 'json_path')
    return construct_single_handler_logger(
        "json-event-record-logger",
        DEBUG,
        JsonEventLoggerHandler(
            json_path,
            lambda record: construct_event_record(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,
                    record=record,
                )
            ),
        ),
    )



import os
from collections import namedtuple
from glob import glob

import six
import yaml

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.utils.yaml_utils import merge_yamls

from .pipeline import PipelineDefinition


class RepositoryDefinition(object):
    '''Define a repository that contains a collection of pipelines.

    Args:
        name (str): The name of the pipeline.
        pipeline_dict (Dict[str, callable]):
            An dictionary of pipelines. The value of the dictionary is a function that takes
            no parameters and returns a PipelineDefiniton.

            We pass callables instead of the PipelineDefinitions itself so that they can be
            created on demand when accessed by name.

            As the pipelines are retrieved it ensures that the keys of the dictionary and the
            name of the pipeline are the same.
        repo_config (Optional[dict]):
            Preset configurations for pipelines such as environments and execution subsets

    '''

    def __init__(self, name, pipeline_dict, repo_config=None, enforce_solid_def_uniqueness=True):
        self.name = check.str_param(name, 'name')

        check.dict_param(pipeline_dict, 'pipeline_dict', key_type=str)

        for val in pipeline_dict.values():
            check.is_callable(val, 'Value in pipeline_dict must be function')

        self.pipeline_dict = pipeline_dict

        self._pipeline_cache = {}

        self.repo_config = check.opt_dict_param(repo_config, 'repo_config')

        self.enforce_solid_def_uniqueness = check.bool_param(
            enforce_solid_def_uniqueness, 'enforce_solid_def_uniqueness'
        )

    def has_pipeline(self, name):
        check.str_param(name, 'name')
        return name in self.pipeline_dict

    def get_pipeline(self, name):
        '''Get a pipeline by name. Only constructs that pipeline and caches it.

        Args:
            name (str): Name of the pipeline to retriever

        Returns:
            PipelineDefinition: Instance of PipelineDefinition with that name.
        '''
        check.str_param(name, 'name')

        if name in self._pipeline_cache:
            return self._pipeline_cache[name]

        try:
            pipeline = self.pipeline_dict[name]()
        except KeyError:
            raise DagsterInvariantViolationError(
                'Could not find pipeline "{name}". Found: {pipeline_names}.'.format(
                    name=name,
                    pipeline_names=', '.join(
                        [
                            '"{pipeline_name}"'.format(pipeline_name=pipeline_name)
                            for pipeline_name in self.pipeline_dict.keys()
                        ]
                    ),
                )
            )
        check.invariant(
            pipeline.name == name,
            'Name does not match. Name in dict {name}. Name in pipeline {pipeline.name}'.format(
                name=name, pipeline=pipeline
            ),
        )

        self._pipeline_cache[name] = check.inst(
            pipeline,
            PipelineDefinition,
            (
                'Function passed into pipeline_dict with key {key} must return a '
                'PipelineDefinition'
            ).format(key=name),
        )

        return pipeline

    def get_all_pipelines(self):
        '''Return all pipelines as a list

        Returns:
            List[PipelineDefinition]:

        '''
        pipelines = list(map(self.get_pipeline, self.pipeline_dict.keys()))
        # This does uniqueness check
        self._construct_solid_defs(pipelines)
        return pipelines

    def _construct_solid_defs(self, pipelines):
        solid_defs = {}
        solid_to_pipeline = {}
        for pipeline in pipelines:
            for solid_def in pipeline.solid_defs:
                if solid_def.name not in solid_defs:
                    solid_defs[solid_def.name] = solid_def
                    solid_to_pipeline[solid_def.name] = pipeline.name

                if (
                    self.enforce_solid_def_uniqueness
                    and not solid_defs[solid_def.name] is solid_def
                ):
                    first_name, second_name = sorted(
                        [solid_to_pipeline[solid_def.name], pipeline.name]
                    )
                    raise DagsterInvalidDefinitionError(
                        (
                            'You have defined two solids named "{solid_def_name}" '
                            'in repository "{repository_name}". Solid names must be '
                            'unique within a repository. The solid has been defined in '
                            'pipeline "{first_pipeline_name}" and it has been defined '
                            'again in pipeline "{second_pipeline_name}."'
                        ).format(
                            solid_def_name=solid_def.name,
                            repository_name=self.name,
                            first_pipeline_name=first_name,
                            second_pipeline_name=second_name,
                        )
                    )

        return solid_defs

    def get_solid_def(self, name):
        check.str_param(name, 'name')

        if not self.enforce_solid_def_uniqueness:
            raise DagsterInvariantViolationError(
                (
                    'Cannot use get_solid_def in repo {} since solid def uniqueness '
                    'is not enforced'
                ).format(self.name)
            )

        solid_defs = self._construct_solid_defs(self.get_all_pipelines())

        if name not in solid_defs:
            check.failed('could not find solid_def {}'.format(name))

        return solid_defs[name]

    def solid_def_named(self, name):
        check.str_param(name, 'name')
        for pipeline in self.get_all_pipelines():
            for solid in pipeline.solids:
                if solid.definition.name == name:
                    return solid.definition

        check.failed('Did not find ' + name)

    def get_preset(self, pipeline_name, solid_name):
        check.str_param(pipeline_name, 'pipeline_name')
        check.str_param(solid_name, 'solid_name')
        return self.get_presets_for_pipeline(pipeline_name).get(solid_name)

    def get_presets_for_pipeline(self, pipeline_name):
        if not (self.repo_config and self.repo_config.get('pipelines')):
            return {}

        presets = self.repo_config['pipelines'].get(pipeline_name, {}).get('presets', {})

        return {
            name: PipelinePreset(
                name, pipeline_name, config.get('solid_subset'), config.get('environment_files')
            )
            for name, config in presets.items()
        }

    def get_preset_pipeline(self, pipeline_name, preset_name):
        pipeline = self.get_pipeline(pipeline_name)
        preset = self.get_preset(pipeline_name, preset_name)
        if not preset:
            preset_names = [name for name in self.get_presets_for_pipeline(pipeline_name)]
            raise DagsterInvariantViolationError(
                'Could not find preset for "{preset_name}". Available presets for pipeline "{pipeline_name}" '
                'are {preset_names}.'.format(
                    preset_name=preset_name, preset_names=preset_names, pipeline_name=pipeline_name
                )
            )

        if not preset.solid_subset is None:
            pipeline = pipeline.build_sub_pipeline(preset.solid_subset)

        return {'pipeline': pipeline, 'environment_dict': preset.environment_dict}


class PipelinePreset(
    namedtuple('_PipelinePreset', 'name pipeline_name solid_subset environment_files')
):
    def __new__(cls, name, pipeline_name, solid_subset=None, environment_files=None):
        return super(PipelinePreset, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(pipeline_name, 'pipeline_name'),
            check.opt_nullable_list_param(solid_subset, 'solid_subset', of_type=str),
            check.opt_nullable_list_param(environment_files, 'environment_files'),
        )

    @property
    def environment_dict(self):
        if self.environment_files is None:
            return None

        file_set = set()
        for file_glob in self.environment_files:
            files = glob(file_glob)
            if not files:
                raise DagsterInvalidDefinitionError(
                    'File or glob pattern "{file_glob}" for "environment_files" in preset "{name}" for '
                    'pipeline "{pipeline_name}" produced no results.'.format(
                        name=self.name, file_glob=file_glob, pipeline_name=self.pipeline_name
                    )
                )

            file_set.update(map(os.path.realpath, files))

        try:
            merged = merge_yamls(list(file_set))
        except yaml.YAMLError as err:
            six.raise_from(
                DagsterInvariantViolationError(
                    'Encountered error attempting to parse yaml. Parsing files {file_set} loaded by '
                    'file/patterns {files} on preset "{name}" for pipeline "{pipeline_name}".'.format(
                        file_set=file_set,
                        files=self.environment_files,
                        name=self.name,
                        pipeline_name=self.pipeline_name,
                    )
                ),
                err,
            )

        return merged

    @property
    def environment_yaml(self):
        merged = self.environment_dict
        if merged is None:
            return None

        return yaml.dump(merged, default_flow_style=False)



from contextlib import contextmanager
from enum import Enum
import sys
import traceback
from future.utils import raise_from


from dagster import check


class DagsterExecutionFailureReason(Enum):
    USER_CODE_ERROR = 'USER_CODE_ERROR'
    FRAMEWORK_ERROR = 'FRAMEWORK_ERROR'
    EXPECTATION_FAILURE = 'EXPECATION_FAILURE'


class DagsterError(Exception):
    @property
    def is_user_code_error(self):
        return False


class DagsterUserError(DagsterError):
    pass


class DagsterRuntimeCoercionError(DagsterError):
    '''Runtime checked faild'''


class DagsterInvalidDefinitionError(DagsterUserError):
    '''Indicates that some violation of the definition rules has been violated by the user'''


class DagsterInvariantViolationError(DagsterUserError):
    '''Indicates the user has violated a well-defined invariant that can only be deteremined
    at runtime.
    '''


class DagsterTypeError(DagsterUserError):
    '''Indicates an error in the solid type system (e.g. mismatched arguments)'''


class DagsterUserCodeExecutionError(DagsterUserError):
    '''
    This is base class for any exception that is meant to wrap an Exception
    thrown by user code. It wraps that existing user code. The original_exc_info
    argument to the ctor is meant to be a sys.exc_info at the site of constructor.
    '''

    def __init__(self, *args, **kwargs):
        # original_exc_info should be gotten from a sys.exc_info() call at the
        # callsite inside of the exception handler. this will allow consuming
        # code to *re-raise* the user error in it's original format
        # for cleaner error reporting that does not have framework code in it
        user_exception = check.inst_param(kwargs.pop('user_exception'), 'user_exception', Exception)
        original_exc_info = check.opt_tuple_param(
            kwargs.pop('original_exc_info', None), 'original_exc_info'
        )

        msg = _add_inner_exception_for_py2(args[0], original_exc_info)

        super(DagsterUserCodeExecutionError, self).__init__(msg, *args[1:], **kwargs)

        self.user_exception = check.opt_inst_param(user_exception, 'user_exception', Exception)
        self.original_exc_info = original_exc_info

    @property
    def is_user_code_error(self):
        return True


class DagsterExecutionStepNotFoundError(DagsterUserError):
    '''
    Generic exception used whenever a user an execution step is specified in an
    api that is consumed by higher level infrastructure (e.g. airflow integration)
    or that needs to be translated into a specific in the GraphQL domain.
    '''

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop('step_key'), 'step_key')
        super(DagsterExecutionStepNotFoundError, self).__init__(*args, **kwargs)


class DagsterExecutionStepExecutionError(DagsterUserCodeExecutionError):
    '''Indicates an error occured during the body of execution step execution'''

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop('step_key'), 'step_key')
        self.solid_name = check.str_param(kwargs.pop('solid_name'), 'solid_name')
        self.solid_def_name = check.str_param(kwargs.pop('solid_def_name'), 'solid_def_name')
        super(DagsterExecutionStepExecutionError, self).__init__(*args, **kwargs)


class DagsterContextFunctionError(DagsterUserCodeExecutionError):
    '''Indicates an error occured during the body of context_fn in a PipelineContextDefinition'''


class DagsterResourceFunctionError(DagsterUserCodeExecutionError):
    '''Indicates an error occured during the body of resource_fn in a ResourceDefinition'''


class DagsterExpectationFailedError(DagsterError):
    '''Thrown with pipeline configured to throw on expectation failure'''

    def __init__(self, expectation_context, value, *args, **kwargs):
        super(DagsterExpectationFailedError, self).__init__(*args, **kwargs)
        self.expectation_context = expectation_context
        self.value = value

    def __repr__(self):
        inout_def = self.expectation_context.inout_def
        return (
            'DagsterExpectationFailedError(solid={solid_name}, {key}={inout_name}, '
            'expectation={e_name}, value={value})'
        ).format(
            solid_name=self.expectation_context.solid.name,
            key=inout_def.descriptive_key,
            inout_name=inout_def.name,
            e_name=self.expectation_context.expectation_def.name,
            value=repr(self.value),
        )

    def __str__(self):
        return self.__repr__()


class DagsterSubprocessExecutionError(DagsterError):
    '''Indicates that an error was encountered when executing part or all of an execution plan
    in a subprocess.
    
    Although this may be a user error, it is intended to be raised in circumstances where we don't
    have access to the sys.exc_info generated by the error (or where the error is raised by a 
    non-Python process).
    '''


class DagsterRunNotFoundError(DagsterError):
    def __init__(self, *args, **kwargs):
        self.invalid_run_id = check.str_param(kwargs.pop('invalid_run_id'), 'invalid_run_id')
        super(DagsterRunNotFoundError, self).__init__(*args, **kwargs)


class DagsterStepOutputNotFoundError(DagsterError):
    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop('step_key'), 'step_key')
        self.output_name = check.str_param(kwargs.pop('output_name'), 'output_name')
        super(DagsterStepOutputNotFoundError, self).__init__(*args, **kwargs)


def _add_inner_exception_for_py2(msg, exc_info):
    if sys.version_info[0] == 2:
        return (
            msg
            + '\n\nThe above exception was the direct cause of the following exception:\n\n'
            + ''.join(traceback.format_exception(*exc_info))
        )

    return msg


@contextmanager
def user_code_error_boundary(error_cls, msg, **kwargs):
    '''
    Wraps the execution of user-space code in an error boundary. This places a uniform
    policy around an user code invoked by the framework. This ensures that all user
    errors are wrapped in the DagsterUserCodeExecutionError, and that the original stack
    trace of the user error is preserved, so that it can be reported without confusing
    framework code in the stack trace, if a tool author wishes to do so. This has
    been especially help in a notebooking context.
    '''
    check.str_param(msg, 'msg')
    check.subclass_param(error_cls, 'error_cls', DagsterUserCodeExecutionError)

    try:
        yield
    except Exception as e:  # pylint: disable=W0703
        if isinstance(e, DagsterError):
            # The system has thrown an error that is part of the user-framework contract
            raise e
        else:
            # An exception has been thrown by user code and computation should cease
            # with the error reported further up the stack
            raise_from(
                error_cls(msg, user_exception=e, original_exc_info=sys.exc_info(), **kwargs), e
            )


class DagsterPY4JTribalKnowledgeException(Exception):
    '''
    The sole purpose of this class is to catalog and note the strange conditions
    under which py4j exceptions are raised. This is just a place to encode
    insitutional knowledge.

    Note: We no longer throw this for now because it is masking errors
    '''


# HAS_P4J = True
# try:
#     import py4j
# except ImportError:
#     HAS_P4J = False

TRIBAL_KNOWLEDGE_ERROR_MESSAGE = '''
Congratulations, you have managed to encountered an error out of Py4J. These
error messages are often quite inscrutable and nearly impossible to act upon
in a meaningful way without serious investigation. As a result, we have
created this generic wrapper exceptions whose sole purpose in life is to document
all the different error conditions that have been reported as py4j errors
in the hopes that others can benefit from these joyous explorations.

Note: The original exception is reraised along with DagsterPY4JTribalKnowledgeException 
in python 3.

Case 1:

An error of the form: 
---------
py4j.protocol.Py4JError: An error occurred while calling o44.__getstate__. Trace:
py4j.Py4JException: Method __getstate__([]) does not exist
        at py4j.reflection.ReflectionEngine.getMethod(ReflectionEngine.java:318)
        at py4j.reflection.ReflectionEngine.getMethod(ReflectionEngine.java:326)
        at py4j.Gateway.invoke(Gateway.java:274)
        at py4j.commands.AbstractCommand.invokeMethod(AbstractCommand.java:132)
        at py4j.commands.CallCommand.execute(CallCommand.java:79)
        at py4j.GatewayConnection.run(GatewayConnection.java:238)
        at java.lang.Thread.run(Thread.java:748)
---------

We have encountered this while attempting to pickle an unpickable object that proxies
into the JVM (such as a spark dataframe).


Original Error Text:
{original_error_text}
'''



from dagster.core.types import Int, Nullable, List
from dagster.core.types.evaluator import validate_config
from dagster.core.types.field import resolve_to_config_type


def test_config_int():
    int_inst = resolve_to_config_type(Int)
    assert not validate_config(int_inst, 1)
    assert validate_config(int_inst, None)
    assert validate_config(int_inst, 'r')

    assert not int_inst.is_list
    assert not int_inst.is_nullable
    assert not (int_inst.is_nullable or int_inst.is_list)


def test_nullable_int():
    nullable_int_inst = resolve_to_config_type(Nullable(Int))
    assert not validate_config(nullable_int_inst, 1)
    assert not validate_config(nullable_int_inst, None)
    assert validate_config(nullable_int_inst, 'r')


def test_list_int():
    list_int = resolve_to_config_type(List(Int))
    assert not validate_config(list_int, [1])
    assert not validate_config(list_int, [1, 2])
    assert not validate_config(list_int, [])
    assert validate_config(list_int, [None])
    assert validate_config(list_int, [1, None])
    assert validate_config(list_int, None)
    assert validate_config(list_int, [1, 'absdf'])


def test_list_nullable_int():
    lni = resolve_to_config_type(List(Nullable(Int)))
    assert not validate_config(lni, [1])
    assert not validate_config(lni, [1, 2])
    assert not validate_config(lni, [])
    assert not validate_config(lni, [None])
    assert not validate_config(lni, [1, None])
    assert validate_config(lni, None)
    assert validate_config(lni, [1, 'absdf'])



class WrappingType(object):
    def __init__(self, inner_type):
        # Cannot check inner_type because of circular references and no fwd declarations
        self.inner_type = inner_type


def List(inner_type):
    '''
    Validates at runtime that the value is ``List[inner_type]``.

    Args:
        inner_type (DagsterType)
    '''
    return WrappingListType(inner_type)


class WrappingListType(WrappingType):
    pass


def Nullable(inner_type):
    '''
    Validates at runtime that the type is either ``None`` or passes validation of ``inner_type``

    Args:
        inner_type (DagsterType)
    '''
    # Cannot check inner_type because of circular references and no fwd declarations
    return WrappingNullableType(inner_type)


class WrappingNullableType(WrappingType):
    pass
